/*
BCH 32.21 extended code soft decision decoding by M-algorithm
Simple and stupid realisation, yet not very computationally efficient

For output BER of 0.001, the soft decision provides a gain of about 2dB
if compared with hard decision techniques for this code.

E/No = 7.0dB (AWGN, BER = 0.001, calculated to usefull data rate)

The reasonable value for M is about 32...64 for this code (ORDER = 5...6)

Input is float array representing input signal readouts (MSFirst)
Expected value for 1 is +1.0, value for 0 is -1.0
Output is maximum likelihood decoded solution.

(c)2000 VLV,  vlv@writeme.com
*/

#define     ORDER                6 // Consider M = pow(2,ORDER) paths
#define     NUMBER_OF_PATHS     (1<<ORDER)


//********** BCH 32.21 encoder *************
// Input is 21-bit data word
// Output is 32-bit BCH 32.21 word

unsigned long int EncodeBCH(unsigned long int x)
{
unsigned char i;
unsigned int p;
unsigned long int xtmp;

xtmp = x;

for(i=0;i<21;i++)
 {
 xtmp<<=1;
 if(xtmp&0x200000) xtmp^=0x1B4800;
 }

xtmp>>=10;
xtmp&=0x7fe;
xtmp|=x<<11;

p=(unsigned int)(xtmp^(xtmp>>16));
p^=p>>8;
p^=p>>4;
p^=p>>2;
p^=p>>1;

xtmp|=p&1;

return xtmp;
}

//***************** Soft Decision Decoder for 32.21 BCH ********
// Input is 32 float values
// Output is 21-bit decoded data word

unsigned long int DecodeBCH(float *x)
{
int i,j,k,l;
unsigned long int utmp;

float ftmp;
float price_of_0,price_of_1;

float min_cost;
int min_tag;

float bit_cost[64];
float path_cost[NUMBER_OF_PATHS];
unsigned long int path[NUMBER_OF_PATHS];

float ext_path_cost[(NUMBER_OF_PATHS<<1)];

float survived_path_cost[NUMBER_OF_PATHS];
unsigned long int survived_path[NUMBER_OF_PATHS];



// Init bit costs (squared error metrics)

j=0;
for(i=0;i<32;i++)
{
ftmp = x[i] + 1.0;                      // Even - cost of 0
bit_cost[j] = ftmp*ftmp;
j++;

ftmp = x[i] - 1.0;                      // Odd - cost of 1
bit_cost[j] = ftmp*ftmp;
j++;
}


//  Initialize paths and costs

for(i=0;i<NUMBER_OF_PATHS;i++)
 {
 path[i]=i;
 path_cost[i]=0.0;
 utmp=i<<1;
 for(j=0;j<ORDER;j++)
  {
  if(utmp&NUMBER_OF_PATHS) path_cost[i]+=bit_cost[(j<<1)+1];
  else path_cost[i]+=bit_cost[(j<<1)];
  utmp<<=1;
  }
 }


// Main loop searching paths

for(i=ORDER;i<21;i++)
 {

 price_of_0 = bit_cost[(i<<1)];
 price_of_1 = bit_cost[(i<<1)+1];

 // Extend the paths

 k=0;
 for(j=0;j<NUMBER_OF_PATHS;j++)
  {
  ext_path_cost[k] = path_cost[j] + price_of_0;
  k++;
  ext_path_cost[k] = path_cost[j] + price_of_1;
  k++;
  }

 // Find the survived paths with minimal costs

 for(j=0;j<NUMBER_OF_PATHS;j++)
  {

  l=0;
  while(ext_path_cost[l]<0.0) l++;
  min_cost = ext_path_cost[l];
  min_tag = l;

  for(k=l+1;k<(NUMBER_OF_PATHS<<1);k++)
   {
   if(ext_path_cost[k]>=0.0)
    {
    if(min_cost > ext_path_cost[k])
     {
     min_cost = ext_path_cost[k];
     min_tag = k;
     }
    }
   }


  ext_path_cost[min_tag] = -1.0;        // Exclude from consideration

  survived_path[j] = (path[min_tag>>1])<<1;
  survived_path[j]|=min_tag&1;
  survived_path_cost[j] = min_cost;
  }

 // Copy survived paths to paths

 for(j=0;j<NUMBER_OF_PATHS;j++)
  {
  path[j]=survived_path[j];
  path_cost[j]=survived_path_cost[j];
  }
 }

//------------------------------------------

 // Encode survived paths and find the full costs

 for(j=0;j<NUMBER_OF_PATHS;j++)
  {
  utmp = EncodeBCH(path[j]);
  ftmp = path_cost[j];

  for(k=0;k<11;k++)
   {
   utmp<<=1;
   if(utmp&0x800) ftmp+=bit_cost[43 + (k<<1)];
   else ftmp+=bit_cost[42 + (k<<1)];
   }
  path_cost[j]=ftmp;
  }

// Find the path with minimum cost

 min_cost = path_cost[0];
 min_tag = 0;

 for(j=1;j<NUMBER_OF_PATHS;j++)
  {
  if(min_cost>path_cost[j])
   {
   min_cost=path_cost[j];
   min_tag = j;
   }
  }
return path[min_tag];
}


